#include "qthing/core.hpp"

#include "qthing/mqtt_ota.hpp"

void qthing::Config::apply() { qthing::getDriver<Core>()->setConfig(*this); };

qthing::Core::Core() {
  add_ota_event_callback([&](ota_event_t event) {
    if (event.state == qthing::start) {
      handleOtaEvent(true, true);
    } else if (event.state == qthing::success || event.state == qthing::error) {
      // After successful ota the esp will be restarted soon, so the ota config can stay active.
      // TODO: check if reboot after ota is configured, if not go back to normal config
      handleOtaEvent(false, event.state == qthing::error);
    }
  });
}
void qthing::Core::handleOtaEvent(bool otaActive, bool applyConfig) {
  const std::lock_guard<std::mutex> lock(configMutex);
  otaStateActive = otaActive;
  if (applyConfig) {
    if (otaConfig) {
      qthing::Core::applyConfig(*otaConfig);
    }
    else {
      qthing::Core::applyConfig(primaryConfig.deriveOtaConfig());
    }
  }
}
void qthing::Core::setConfig(const qthing::Config& config) {
  const std::lock_guard<std::mutex> lock(configMutex);
  primaryConfig = config;
  if (!otaStateActive) {
    // no ota, config changed: apply config
    qthing::Core::applyConfig(primaryConfig);
  } else {
    // ota, config changed: derive ota config is there is no explicit ota config configured
    if (!otaConfig) {
      qthing::Core::applyConfig(primaryConfig.deriveOtaConfig());
    }
  }
}
void qthing::Core::setOtaConfig(const std::optional<qthing::Config>& newOtaConfig) {
  const std::lock_guard<std::mutex> lock(configMutex);
  otaConfig = newOtaConfig;
  if (otaStateActive) {
    if (otaConfig) {
      qthing::Core::applyConfig(*otaConfig);
    }
    else {
      qthing::Core::applyConfig(primaryConfig.deriveOtaConfig());
    }
  }
}
void qthing::Core::applyConfig(const qthing::Config& config) {
  std::unordered_set<std::shared_ptr<Driver>> changedDrivers;
  // Remove components no longer present in config
  for (auto&& oldComponent : activeConfig.components()) {
    if (config.components().find(oldComponent) == config.components().end()) {
      oldComponent->deinit();
      changedDrivers.insert(oldComponent->driver());
    }
  }
  // Add new components
  for (auto&& newComponent : config.components()) {
    if (activeConfig.components().find(newComponent) == activeConfig.components().end()) {
      newComponent->init();
      changedDrivers.insert(newComponent->driver());
    }
  }
  for (auto&& driver : changedDrivers) {
    if (driver != NULL) {
      driver->update();
    }
  }
  activeConfig = config;
  onActiveConfigChanged(config);
}
