#pragma once

#include <sys/types.h>

#include <cmath>

namespace qthing {

struct RGBW8 {
  union {
    struct __attribute__((packed)) {
      uint8_t r, g, b, w;
    };
    uint32_t num;
  };
  RGBW8(u_int8_t r, uint8_t g, uint8_t b, uint8_t w) : r(r), g(g), b(b), w(w) {}
  RGBW8() : num(0) {}
  RGBW8& operator+=(const RGBW8& other) {
    this->r += other.r;
    this->g += other.g;
    this->b += other.b;
    this->w += other.w;
    return *this;
  }
  RGBW8& operator*=(float scale) {
    this->r *= scale;
    this->g *= scale;
    this->b *= scale;
    this->w *= scale;
    return *this;
  }
};
inline RGBW8 operator+(RGBW8 lhs, const RGBW8& rhs) { return lhs += rhs; }
inline RGBW8 operator*(RGBW8 lhs, float rhs) { return lhs *= rhs; }
inline RGBW8 operator*(float lhs, RGBW8 rhs) { return rhs *= lhs; }

struct RGBW {
  float r;
  float g;
  float b;
  float w;
  RGBW() : r(0), g(0), b(0), w(0) {}
  RGBW(float r, float g, float b, float w) : r(r), g(g), b(b), w(w) {}
  RGBW& operator+=(const RGBW& other) {
    this->r += other.r;
    this->g += other.g;
    this->b += other.b;
    this->w += other.w;
    return *this;
  }
  RGBW& operator*=(float scale) {
    this->r *= scale;
    this->g *= scale;
    this->b *= scale;
    this->w *= scale;
    return *this;
  }
  explicit operator RGBW8() { return RGBW8(this->r * 255, this->g * 255, this->b * 255, this->w * 255); }
};
inline RGBW operator+(RGBW lhs, const RGBW& rhs) { return lhs += rhs; }
inline RGBW operator*(RGBW lhs, float rhs) { return lhs *= rhs; }
inline RGBW operator*(float lhs, RGBW rhs) { return rhs *= lhs; }

struct RGB {
  float r;
  float g;
  float b;
  RGB() : r(0), g(0), b(0) {}
  RGB(float r, float g, float b) : r(r), g(g), b(b) {}
  RGB& operator+=(const RGB& other) {
    this->r += other.r;
    this->g += other.g;
    this->b += other.b;
    return *this;
  }
  RGB& operator*=(float scale) {
    this->r *= scale;
    this->g *= scale;
    this->b *= scale;
    return *this;
  }
  operator RGBW() const { return RGBW(this->r, this->g, this->b, 0); }
  explicit operator RGBW8() { return RGBW8(this->r * 255, this->g * 255, this->b * 255, 0); }
};
inline RGB operator+(RGB lhs, const RGB& rhs) { return lhs += rhs; }
inline RGB operator*(RGB lhs, float rhs) { return lhs *= rhs; }
inline RGB operator*(float lhs, RGB rhs) { return rhs *= lhs; }

struct GammaCorrection {
  float gamma_r;
  float gamma_g;
  float gamma_b;
  float gamma_w;
  float max_r;
  float max_g;
  float max_b;
  float max_w;
  GammaCorrection(const float gamma_r, const float gamma_g, const float gamma_b, const float gamma_w, const float max_r,
                  const float max_g, const float max_b, const float max_w)
      : gamma_r(gamma_r),
        gamma_g(gamma_g),
        gamma_b(gamma_b),
        gamma_w(gamma_w),
        max_r(max_r),
        max_g(max_g),
        max_b(max_b),
        max_w(max_w) {}
  GammaCorrection(const float gamma_r, const float gamma_g, const float gamma_b, const float gamma_w)
      : GammaCorrection(gamma_r, gamma_g, gamma_b, gamma_w, 1, 1, 1, 1) {}
  GammaCorrection(const float gamma) : GammaCorrection(gamma, gamma, gamma, gamma) {}
  const RGBW operator()(const RGBW& color) {
    return RGBW(pow(color.r, gamma_r) * max_r, pow(color.g, gamma_g) * max_g, pow(color.b, gamma_b) * max_b,
                pow(color.w, gamma_w) * max_w);
  }
};

RGB hue(const float hue, const float value = 1);
RGB hsv(float hue, float staturation = 1, float value = 1);

struct HSV {
  float hue;
  float saturation;
  float value;
  HSV(float hue, float saturation = 1, float value = 1) : hue(hue), saturation(saturation), value(value) {}
  HSV& operator*=(float scale) {
    value *= scale;
    return *this;
  }
  explicit operator RGB() { return hsv(hue, saturation, value); }
  explicit operator RGBW() { return static_cast<RGB>(*this); }
  explicit operator RGBW8() { return static_cast<RGBW8>(static_cast<RGB>(*this)); }
};
inline HSV operator*(HSV lhs, float rhs) { return lhs *= rhs; }
inline HSV operator*(float lhs, HSV rhs) { return rhs *= lhs; }

}  // namespace qthing
