#pragma once

#include <algorithm>
#include <functional>
#include <memory>
#include <mutex>
#include <optional>
#include <type_traits>
#include <unordered_map>
#include <unordered_set>
#include <vector>

#include "qthing/callback.hpp"
#include "qthing/description.hpp"

namespace qthing {

class Driver {
  friend class Core;

  Driver(const Driver&) = delete;             // delete cctor
  Driver& operator=(const Driver&) = delete;  // delete cassop
  Driver(Driver&&) = delete;                  // delete mctor
  Driver& operator=(Driver&&) = delete;       // delete massop

 protected:
  Driver() {}
  virtual void update(){};

 public:
  virtual ~Driver() {}
};

template <typename T>
class DriverStorage {
  static_assert(std::is_convertible<T*, Driver*>::value,
                "`getDriver` can only be used with types that are deriving `Driver`");
  DriverStorage() = delete;                                       // delete ctor (static class)
  DriverStorage(const DriverStorage<T>&) = delete;                // delete cctor
  DriverStorage<T>& operator=(const DriverStorage<T>&) = delete;  // delete cassop
  DriverStorage(DriverStorage<T>&&) = delete;                     // delete mctor
  DriverStorage<T>& operator=(DriverStorage<T>&&) = delete;       // delete massop

 public:
  static std::shared_ptr<T> get() {
    static T instance;
    static std::shared_ptr<T> instancePointer(&instance);
    return instancePointer;
  }
  static std::shared_ptr<Driver> getDriver() { return std::static_pointer_cast<Driver>(get()); }
};

template <typename T>
std::shared_ptr<T> getDriver() {
  return DriverStorage<T>::get();
}

class Component : public std::enable_shared_from_this<Component> {
  friend class Core;
  Component(const Component&) = delete;             // delete cctor
  Component& operator=(const Component&) = delete;  // delete cassop
  Component(Component&&) = delete;                  // delete mctor
  Component& operator=(Component&&) = delete;       // delete massop

 protected:
  Component() {}
  virtual void init() = 0;
  virtual void deinit() = 0;
  virtual std::shared_ptr<Driver> driver() { return NULL; };

 public:
  virtual ~Component() {}
  virtual void describe(DeviceDescription& description) = 0;
  virtual std::shared_ptr<Component> deriveOtaComponent() { return this->shared_from_this(); }
};

template <class TSelf, typename TDriver>
class ComponentWithDriver : public Component {
  void init() final override { getDriver<TDriver>()->add(std::static_pointer_cast<TSelf>(this->shared_from_this())); };
  void deinit() final override {
    getDriver<TDriver>()->remove(std::static_pointer_cast<TSelf>(this->shared_from_this()));
  }
  std::shared_ptr<Driver> driver() final override { return DriverStorage<TDriver>::getDriver(); }
};

class Config {
 private:
  std::unordered_set<std::shared_ptr<Component>> _components;

 public:
  Config() {}
  Config(std::initializer_list<std::shared_ptr<Component>> components) : _components(components){};
  Config(std::unordered_set<std::shared_ptr<Component>> components) : _components(components) {}
  const std::unordered_set<std::shared_ptr<Component>>& components() const { return _components; }
  template <typename T, typename... Args>
  std::shared_ptr<T> add(Args... args) {
    static_assert(std::is_convertible<T*, Component*>::value,
                  "`add` can only be used with types that are deriving `Component`");
    auto component = std::make_shared<T>(args...);
    auto [_, success] = _components.insert(component);
    if (!success) {
      throw std::runtime_error("Could not add component");
    }
    return component;
  }
  void remove(std::shared_ptr<Component> component) { _components.erase(component); }
  void apply();
  Config deriveOtaConfig() {
    std::unordered_set<std::shared_ptr<Component>> otaComponents;
    for (auto component : _components) {
      auto otaComponent = component->deriveOtaComponent();
      if (otaComponent) {
        otaComponents.insert(otaComponent);
      }
    }
    return Config(otaComponents);
  }
};

class Core : public Driver {
  friend DriverStorage<Core>;

  Core(const Core&) = delete;             // delete cctor
  Core& operator=(const Core&) = delete;  // delete cassop
  Core(Core&&) = delete;                  // delete mctor
  Core& operator=(Core&&) = delete;       // delete massop

 private:
  std::mutex configMutex;
  bool otaStateActive;
  Config primaryConfig;
  std::optional<Config> otaConfig;
  Config activeConfig;

  Core();  // declare ctor private
  void handleOtaEvent(bool otaActive, bool applyConfig);
  void applyConfig(const Config& config);

 public:
  const Callback<void(const Config&), Core> onActiveConfigChanged;

  const Config& getActiveConfig() { return activeConfig; }
  void setConfig(const Config& config);
  const Config& getConfig() { return primaryConfig; }
  void setOtaConfig(const std::optional<Config>& otaConfig);
  const std::optional<Config>& getOtaConfig() { return otaConfig; }
};

}  // namespace qthing
