#pragma once

#include <sys/_stdint.h>

#include <stdexcept>

#include "driver/gpio.h"
#include "qthing/color.hpp"
#include "qthing/core.hpp"
#include "qthing/ota.hpp"

namespace qthing::leds {

namespace internal {
class RmtLedStripDriver;  // forward declaration
}

enum class RmtLedStripType {
  Invalid,
  WS2812_V1,
  WS2812B_V1,
  WS2812B_V2,
  WS2812B_V3,
  WS2813_V1,
  WS2813_V2,
  WS2813_V3,
  SK6812_V1,
  SK6812W_V1,
};

class RmtLedStrip : public ComponentWithDriver<RmtLedStrip, internal::RmtLedStripDriver> {
  RmtLedStrip(const RmtLedStrip&) = delete;             // delete cctor
  RmtLedStrip& operator=(const RmtLedStrip&) = delete;  // delete cassop
  RmtLedStrip(RmtLedStrip&&) = delete;                  // delete mctor
  RmtLedStrip& operator=(RmtLedStrip&&) = delete;       // delete massop

 public:
  const RmtLedStripType ledStripType;
  const gpio_num_t gpio;
  const uint16_t numLeds;
  const OtaLedBehavior otaBehavior;
  std::function<qthing::RGBW8(uint16_t)> colorHandler;
  RmtLedStrip(RmtLedStripType ledStripType, gpio_num_t gpio, uint16_t numLeds,
              OtaLedBehavior otaBehavior = OtaLedBehavior::otaStatus)
      : ledStripType(ledStripType), gpio(gpio), numLeds(numLeds), otaBehavior(otaBehavior) {}
  void describe(DeviceDescription& description) override { description.addLeds(numLeds); }
  std::shared_ptr<Component> deriveOtaComponent() override {
    switch (otaBehavior) {
      case OtaLedBehavior::otaStatus: {
        auto otaStrip = std::make_shared<RmtLedStrip>(ledStripType, gpio, 1);
        otaStrip->colorHandler = getDriver<OtaDriver>()->otaColorHandler();
        return otaStrip;
      }
      case OtaLedBehavior::off:
        return NULL;
      case OtaLedBehavior::normal:
        return this->shared_from_this();
      default:
        throw new std::out_of_range("Ota led behavior not supported");
    }
  }
};

namespace internal {

template <RmtLedStripType templateRmtLedStripType>
class RmtLedStripTemplate : public RmtLedStrip {
  RmtLedStripTemplate(const RmtLedStripTemplate&) = delete;             // delete cctor
  RmtLedStripTemplate& operator=(const RmtLedStripTemplate&) = delete;  // delete cassop
  RmtLedStripTemplate(RmtLedStripTemplate&&) = delete;                  // delete mctor
  RmtLedStripTemplate& operator=(RmtLedStripTemplate&&) = delete;       // delete massop
 public:
  RmtLedStripTemplate(gpio_num_t gpio, int numLeds, OtaLedBehavior otaBehavior = OtaLedBehavior::otaStatus) : RmtLedStrip(templateRmtLedStripType, gpio, numLeds, otaBehavior) {}
};

class RmtLedStripDriver : public Driver {
  friend DriverStorage<RmtLedStripDriver>;

  RmtLedStripDriver(const RmtLedStripDriver&) = delete;             // delete cctor
  RmtLedStripDriver& operator=(const RmtLedStripDriver&) = delete;  // delete cassop
  RmtLedStripDriver(RmtLedStripDriver&&) = delete;                  // delete mctor
  RmtLedStripDriver& operator=(RmtLedStripDriver&&) = delete;       // delete massop

 private:
  RmtLedStripDriver();  // declare ctor private
  std::unordered_set<std::shared_ptr<RmtLedStrip>> ledStrips;

 public:
  void add(std::shared_ptr<RmtLedStrip> strip);
  void remove(std::shared_ptr<RmtLedStrip> strip);
  void update() override final;
};

}  // namespace internal

using WS2812_V1 = internal::RmtLedStripTemplate<RmtLedStripType::WS2812_V1>;
using WS2812B_V1 = internal::RmtLedStripTemplate<RmtLedStripType::WS2812B_V1>;
using WS2812B_V2 = internal::RmtLedStripTemplate<RmtLedStripType::WS2812B_V2>;
using WS2812B_V3 = internal::RmtLedStripTemplate<RmtLedStripType::WS2812B_V3>;
using WS2813_V1 = internal::RmtLedStripTemplate<RmtLedStripType::WS2813_V1>;
using WS2813_V2 = internal::RmtLedStripTemplate<RmtLedStripType::WS2813_V2>;
using WS2813_V3 = internal::RmtLedStripTemplate<RmtLedStripType::WS2813_V3>;
using SK6812_V1 = internal::RmtLedStripTemplate<RmtLedStripType::SK6812_V1>;
using SK6812W_V1 = internal::RmtLedStripTemplate<RmtLedStripType::SK6812W_V1>;

}  // namespace qthing::leds
