#include "qthing/leds/led_strip.hpp"

#include <driver/gpio.h>
#include <esp32_digital_led_lib.h>
#include <esp_system.h>
#include <freertos/FreeRTOS.h>
#include <freertos/task.h>
#include <stdlib.h>
#include <sys/_stdint.h>

#include <future>

#include "qthing/color.hpp"
#include "qthing/power.hpp"
#include "qthing/util.hpp"

static uint8_t renderFrequency = 50;

static led_types translateLedType(qthing::leds::RmtLedStripType ledType) {
  switch (ledType) {
    case qthing::leds::RmtLedStripType::WS2812_V1:
      return led_types::LED_WS2812_V1;
    case qthing::leds::RmtLedStripType::WS2812B_V1:
      return led_types::LED_WS2812B_V1;
    case qthing::leds::RmtLedStripType::WS2812B_V2:
      return led_types::LED_WS2812B_V2;
    case qthing::leds::RmtLedStripType::WS2812B_V3:
      return led_types::LED_WS2812B_V3;
    case qthing::leds::RmtLedStripType::WS2813_V1:
      return led_types::LED_WS2813_V1;
    case qthing::leds::RmtLedStripType::WS2813_V2:
      return led_types::LED_WS2813_V2;
    case qthing::leds::RmtLedStripType::WS2813_V3:
      return led_types::LED_WS2813_V3;
    case qthing::leds::RmtLedStripType::SK6812_V1:
      return led_types::LED_SK6812_V1;
    case qthing::leds::RmtLedStripType::SK6812W_V1:
      return led_types::LED_SK6812W_V1;
    default:
      throw std::invalid_argument("ledType: Led strip type not supported");
  }
}

static std::promise<void> renderLoopCompletedPromise;
static volatile bool renderLoopRunning;
static volatile bool renderLoopRequestCancellation;

static size_t count;
static std::array<strand_t, 8> strands;
static std::array<strand_t*, 8> strandAddresses;
static std::array<std::shared_ptr<qthing::leds::RmtLedStrip>, 8> strips;

static void renderLoop(void* ignored) {
  TickType_t lastWakeTime = xTaskGetTickCount();

  digitalLeds_initDriver();
  digitalLeds_addStrands(strandAddresses.data(), count);
  while (!renderLoopRequestCancellation) {
    for (uint8_t i = 0; i < count; i++) {
      auto colorHandler = strips[i]->colorHandler;
      for (uint16_t led = 0; led < strands[i].numPixels; led++) {
        qthing::RGBW8 pixel = colorHandler == NULL ? qthing::RGBW8() : colorHandler(led);
        strands[i].pixels[led] = pixelFromRGBW(pixel.r, pixel.g, pixel.b, pixel.w);
      }
    }
    digitalLeds_drawPixels(strandAddresses.data(), count);
    vTaskDelayUntil(&lastWakeTime, 1000 / renderFrequency / portTICK_PERIOD_MS);
  }
  // TODO: keep last color if requested
  digitalLeds_resetPixels(strandAddresses.data(), count);
  digitalLeds_removeStrands(strandAddresses.data(), count);

  renderLoopCompletedPromise.set_value();
  vTaskDelete(NULL);
}

static void shutdownHandler() {
  for (size_t i = 0; i < 8; i += 1) {
    gpio_reset_pin(static_cast<gpio_num_t>(strands[i].gpioNum));
  }
}

qthing::leds::internal::RmtLedStripDriver::RmtLedStripDriver() {
  esp_register_shutdown_handler(shutdownHandler);

  std::transform(strands.begin(), strands.end(), strandAddresses.begin(), [](strand_t& strand) { return &strand; });
}
void qthing::leds::internal::RmtLedStripDriver::add(std::shared_ptr<RmtLedStrip> strip) { ledStrips.insert(strip); }
void qthing::leds::internal::RmtLedStripDriver::remove(std::shared_ptr<RmtLedStrip> strip) { ledStrips.erase(strip); }
void qthing::leds::internal::RmtLedStripDriver::update() {
  qthing::power_managment_max_power();

  if (renderLoopRunning) {
    // stop render loop
    renderLoopRequestCancellation = true;
    renderLoopCompletedPromise.get_future().wait();

    // cleanup
    for (int i = 0; i < count; i += 1) {
      strips[i] = NULL;
    }
  }

  count = ledStrips.size();
  int i = 0;
  for (auto&& ledStrip : ledStrips) {
    strands[i] = {
        .rmtChannel = i,
        .gpioNum = ledStrip->gpio,
        .ledType = translateLedType(ledStrip->ledStripType),
        .brightLimit = 255,
        .numPixels = ledStrip->numLeds,
        .pixels = nullptr,
        ._stateVars = nullptr,
    };
    strips[i] = ledStrip;
    i += 1;
  }

  renderLoopCompletedPromise = std::promise<void>();
  renderLoopRequestCancellation = false;
  renderLoopRunning = true;

  xTaskCreatePinnedToCore(renderLoop, "renderLoop", 10000, NULL, 6, NULL, 1);
};
