#include <stdio.h>
#include <string.h>

#include "device_config.h"
#include "driver/gpio.h"
#include "environment.h"
#include "esp_eth.h"
#include "esp_event.h"
#include "esp_log.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "qthing/error.hpp"
#include "qthing/network.hpp"
#include "tcpip_adapter.h"

static const char* TAG = "ethernet";

// RMII data pins are fixed:
// TXD0 = GPIO19
// TXD1 = GPIO22
// TX_EN = GPIO21
// RXD0 = GPIO25
// RXD1 = GPIO26
// CLK == GPIO0

constexpr gpio_num_t PIN_PHY_POWER = GPIO_NUM_18;
constexpr gpio_num_t PIN_SMI_MDC = GPIO_NUM_16;
constexpr gpio_num_t PIN_SMI_MDIO = GPIO_NUM_17;

/** Event handler for Ethernet events */
static void eth_event_handler(void* arg, esp_event_base_t event_base, int32_t event_id, void* event_data) {
  uint8_t mac_addr[6] = {0};
  esp_eth_handle_t eth_handle = *(esp_eth_handle_t*)event_data;

  switch (event_id) {
    case ETHERNET_EVENT_CONNECTED:
      esp_eth_ioctl(eth_handle, ETH_CMD_G_MAC_ADDR, mac_addr);
      ESP_LOGI(TAG, "Ethernet Link Up");
      ESP_LOGI(TAG, "Ethernet MAC address %02x:%02x:%02x:%02x:%02x:%02x", mac_addr[0], mac_addr[1], mac_addr[2],
               mac_addr[3], mac_addr[4], mac_addr[5]);
      break;
    default:
      break;
  }
}

void qthing::enable_ethernet_lan8720() {
  qthing::initialize_network();

  // Enable phy clock
  gpio_pad_select_gpio(PIN_PHY_POWER);
  gpio_set_direction(PIN_PHY_POWER, GPIO_MODE_OUTPUT);
  gpio_set_level(PIN_PHY_POWER, 1);
  // Allow the power up/down to take effect by waiting two ticks (one tick should suffice for gpio changes but showed to
  // be problematic)
  vTaskDelay(2);

  try {
    // Not required here, happens in qthing::initialize_network();
    // ESP_ERROR_CHECK(esp_event_loop_create_default());
    errorCodeToException(tcpip_adapter_set_default_eth_handlers());
    errorCodeToException(esp_event_handler_register(ETH_EVENT, ESP_EVENT_ANY_ID, &eth_event_handler, NULL));

#ifdef ETH_STATIC_IP
#ifdef ETH_IP
    tcpip_adapter_dhcpc_stop(TCPIP_ADAPTER_IF_ETH);
    tcpip_adapter_ip_info_t ipInfo;
    inet_pton(AF_INET, ETH_IP, &ipInfo.ip);
    inet_pton(AF_INET, ETH_GATEWAY, &ipInfo.gw);
    inet_pton(AF_INET, ETH_NETMASK, &ipInfo.netmask);
    tcpip_adapter_set_ip_info(TCPIP_ADAPTER_IF_ETH, &ipInfo);
#endif
#endif

    eth_mac_config_t mac_config = ETH_MAC_DEFAULT_CONFIG();
    eth_phy_config_t phy_config = ETH_PHY_DEFAULT_CONFIG();
    phy_config.phy_addr = 1;
    phy_config.reset_gpio_num = -1;
    mac_config.smi_mdc_gpio_num = PIN_SMI_MDC;
    mac_config.smi_mdio_gpio_num = PIN_SMI_MDIO;
    esp_eth_mac_t* mac = esp_eth_mac_new_esp32(&mac_config);
    esp_eth_phy_t* phy = esp_eth_phy_new_lan8720(&phy_config);

    esp_eth_config_t config = ETH_DEFAULT_CONFIG(mac, phy);
    esp_eth_handle_t eth_handle = NULL;
    errorCodeToException(esp_eth_driver_install(&config, &eth_handle));
    /* start Ethernet driver state machine */
    errorCodeToException(esp_eth_start(eth_handle));

  } catch (EspError& err) {
    ESP_LOGE(TAG, "Ethernet could not be initialized (%s)", err.what());
  }
}
